"use strict";

var _interopRequireWildcard = require("babel-runtime/helpers/interop-require-wildcard")["default"];

exports.__esModule = true;
exports._params = _params;
exports._functionBody = _functionBody;
exports._method = _method;
exports.FunctionExpression = FunctionExpression;
exports.FunctionDeclaration = FunctionDeclaration;
exports._finishFunction = _finishFunction;
exports.ArrowFunctionExpression = ArrowFunctionExpression;

var _babelTypes = require("babel-types");

var t = _interopRequireWildcard(_babelTypes);

// TODO: create a "horchata-types" and switch to it

function _params(node) {
  var _this = this;

  if (node.typeParameters) this.print(node, "typeParameters");
  this.printArguments(node, "params", {
    iterator: function iterator(node) {
      // for flow
      if (node.optional) _this.push("?");
      if (node.typeAnnotation) _this.print(node, "typeAnnotation");
    }
  });
}

/**
 * Prints a function body, like printBlock
 */

function _functionBody(parent) {
  var prop = arguments.length <= 1 || arguments[1] === undefined ? "body" : arguments[1];

  var node = parent[prop];
  var opts = {};
  if (t.isBlock(node)) {
    this.indent();
    if (node[prop].length + (node.directives == null ? 0 : node.directives.length) > 0) {
      this.newline();
    }
    this._startPrint(parent, prop, opts);
    if (node.directives) this.printStatements(node, 'directives', opts);
    this.printStatements(node, prop, opts);
    this._finishPrint(node, parent, opts);
    this.dedent();
    // } else if (t.isEmptyStatement(node)) {
    //   // probably not needed
    //   this.push({type: 'pass', after: [';', '\n']}, parent, prop, opts);
  } else {
      throw new Error("Invalid Function body");
    }
  // TODO: transform to tacoscript nodes for implicit return forms
}

/**
 * Prints method-like nodes, prints key, value, and body, handles async, generator, computed, and get or set.
 */

function _method(node, isValue) {
  var value = isValue ? node : node.value;
  var kind = node.kind;

  if (kind === "get" || kind === "set") {
    this.push(kind);
  }

  if (node.computed) {
    this.push("[");
    this.print(node, "key");
    this.push("]");
  } else {
    this.print(node, "key");
  }

  this._finishFunction(value, node);
}

/**
 * Prints FunctionExpression, prints id and body, handles async and generator.
 */

function FunctionExpression(node, parent) {

  var needsParens = t.isCallExpression(parent) && node === parent.callee; // TODO: better needsParens testing
  if (needsParens) this.push("(");

  if (node.id) {
    this.push("function");
    this.print(node, "id");
  }
  this._finishFunction(node, parent);
  if (needsParens) this.push(")");
}

/**
 * Prints FunctionDeclaration, prints id and body, handles async and generator.
 */

function FunctionDeclaration(node, parent) {
  this.push("function");
  if (node.id) {
    this.print(node, "id");
  }
  this._finishFunction(node, parent);
  this.newline(); // note: won't be printed if there's already a newline
}

function _finishFunction(node, parent) {
  this._params(node);

  if (node.generator) this.push("*");
  if (node.async) {
    this.push("+>");
  } else {
    this.push("->");
  }
  this._functionBody(node);
}

/**
 * Prints ArrowFunctionExpression, prints params and body, handles async.
 * TODO: Preserve if parens were included for a single argument
 */

function ArrowFunctionExpression(node, parent) {
  this._params(node);

  if (t.isBlock(node.body)) {
    if (node.async) {
      this.push("+=>");
    } else {
      this.push("=>");
    }
    this._functionBody(node);
  } else {
    if (node.async) {
      this.push("+=>>");
    } else {
      this.push("=>>");
    }
    this.print(node, "body");
  }
}

// TODO: SharpArrowFunctionExpression
// TODO: SharpFunctionExpression
// TODO: SharpFunctionDeclaration